#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "nlltypes.h"
#include "nlllib.h"
#include "memory.h"
#include "value.h"
#include "array.h"

static int array_num = 0;

#ifndef NLL_MEMORY_STATIC
static int _init(void)
{
  return 0;
}

static int _check(void)
{
  return 0;
}

static array_t _alloc(int number)
{
  array_t array;
  if ((array = memory_alloc(sizeof(*array) * number)) == NULL)
    return NULL;
  array_num += number;
  return array;
}

static int _free(array_t array, int number)
{
  array_num -= number;
  memory_free(array);
  return 0;
}
#else
static struct array arrays[ARRAY_MAXNUM];

static int _init(void)
{
  memset(arrays, 0, sizeof(arrays));
  return 0;
}

static int _check(void)
{
  return 0;
}

static array_t _alloc(int number)
{
  array_t array;

  if (array_num + number > ARRAY_MAXNUM)
    return NULL;

  array = &arrays[array_num];
  array_num += number;

  return array;
}

static int _free(array_t array, int number)
{
  while ((array_num > 0) && !arrays[array_num - 1].value)
    array_num--;

  return 0;
}
#endif

int array_check(void)
{
  if (array_num)
    return NLL_ERRCODE_ARRAY_NOT_EMPTY;

  return _check();
}

int array_init(void)
{
  array_num = 0;
  return _init();
}

int array_get(array_t *arrayp, int number)
{
  array_t array;
  int r, i;

  if ((array = _alloc(number)) == NULL)
    return NLL_ERRCODE_ARRAY_BUFFER_OVER;

  memset(array, 0, sizeof(*array) * number);

  for (i = 0; i < number; i++) {
    if ((r = value_alloc(&array[i].value)) < 0) {
      array_free(array, number);
      return r;
    }
  }

  *arrayp = array;

  return 0;
}

int array_free(array_t array, int number)
{
  int r, i;

  for (i = 0; i < number; i++) {
    if (array[i].value) {
      value_free(array[i].value);
      array[i].value = NULL;
    }
  }

  if ((r = _free(array, number)) < 0)
    return r;

  return 0;
}
