#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#ifdef USE_MATHLIB
#include <math.h>
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "const.h"
#include "nlltypes.h"
#include "nlllib.h"
#include "value.h"
#include "area.h"
#include "stack.h"
#include "line.h"
#include "position.h"
#include "nll.h"
#include "function.h"
#include "function-ext.h"
#include "libprintf.h"
#include "key.h"
#include "mouse.h"
#include "framebuf.h"
#include "graphic.h"
#include "audio.h"

static int function_fargs(value_t *valuep, value_t args);
static int function_copy(value_t *valuep, value_t args);
static int function_true(value_t *valuep, value_t args);
static int function_false(value_t *valuep, value_t args);
static int function_cond(value_t *valuep, value_t args);
static int function_at(value_t *valuep, value_t args);
static int function_fpush(value_t *valuep, value_t args);
static int function_fpop(value_t *valuep, value_t args);
static int function_stacknum(value_t *valuep, value_t args);
static int function_fprint(value_t *valuep, value_t args);
static int function_sprintf(value_t *valuep, value_t args);
static int function_fprintf(value_t *valuep, value_t args);
static int function_fprintm(value_t *valuep, value_t args);
static int function_input(value_t *valuep, value_t args);
static int function_fwait(value_t *valuep, value_t args);
static int function_nosystem(value_t *valuep, value_t args);
static int function_nosyscall(value_t *valuep, value_t args);
#ifdef NLL_STDIO_FUNCTION
static int function_fstdin(value_t *valuep, value_t args);
static int function_fstdout(value_t *valuep, value_t args);
#endif
#ifdef NLL_SYSTEM_FUNCTION
static int function_system(value_t *valuep, value_t args);
#endif
#ifdef NLL_SYSCALL_FUNCTION
static int function_syscall(value_t *valuep, value_t args);
#endif
static int function_feval(value_t *valuep, value_t args);
static int function_itoc(value_t *valuep, value_t args);
static int function_ctoi(value_t *valuep, value_t args);
static int function_itoa(value_t *valuep, value_t args);
static int function_atoi(value_t *valuep, value_t args);
#ifdef NLL_FLOATING_POINT
static int function_itof(value_t *valuep, value_t args);
static int function_ftoi(value_t *valuep, value_t args);
#endif
static int function_length(value_t *valuep, value_t args);
static int function_left(value_t *valuep, value_t args);
static int function_right(value_t *valuep, value_t args);
static int function_mid(value_t *valuep, value_t args);
static int function_instr(value_t *valuep, value_t args);
static int function_strcpy(value_t *valuep, value_t args);
static int function_srand(value_t *valuep, value_t args);
static int function_rand(value_t *valuep, value_t args);
static int function_int(value_t *valuep, value_t args);
static int function_abs(value_t *valuep, value_t args);
static int function_sign(value_t *valuep, value_t args);
static int function_time(value_t *valuep, value_t args);
static int function_date(value_t *valuep, value_t args);
static int function_memory(value_t *valuep, value_t args);
static int function_peek(value_t *valuep, value_t args);
static int function_poke(value_t *valuep, value_t args);
static int function_read(value_t *valuep, value_t args);
static int function_write(value_t *valuep, value_t args);
static int function_memcpy(value_t *valuep, value_t args);
#ifdef USE_MATHLIB
static int function_pi(value_t *valuep, value_t args);
static int function_radian(value_t *valuep, value_t args);
static int function_degree(value_t *valuep, value_t args);
static int function_sin(value_t *valuep, value_t args);
static int function_cos(value_t *valuep, value_t args);
static int function_tan(value_t *valuep, value_t args);
static int function_asin(value_t *valuep, value_t args);
static int function_acos(value_t *valuep, value_t args);
static int function_atan(value_t *valuep, value_t args);
static int function_sinh(value_t *valuep, value_t args);
static int function_cosh(value_t *valuep, value_t args);
static int function_tanh(value_t *valuep, value_t args);
static int function_exp(value_t *valuep, value_t args);
static int function_log(value_t *valuep, value_t args);
static int function_log10(value_t *valuep, value_t args);
static int function_sqrt(value_t *valuep, value_t args);
static int function_atan2(value_t *valuep, value_t args);
static int function_pow(value_t *valuep, value_t args);
static int function_fmod(value_t *valuep, value_t args);
static int function_infinity(value_t *valuep, value_t args);
static int function_nan(value_t *valuep, value_t args);
static int function_isfinite(value_t *valuep, value_t args);
static int function_isinf(value_t *valuep, value_t args);
static int function_isnan(value_t *valuep, value_t args);
static int function_isnormal(value_t *valuep, value_t args);
#endif
static int function_fopen(value_t *valuep, value_t args);
static int function_fclose(value_t *valuep, value_t args);
static int function_fgets(value_t *valuep, value_t args);
static int function_fputs(value_t *valuep, value_t args);
static int function_fread(value_t *valuep, value_t args);
static int function_fwrite(value_t *valuep, value_t args);
static int function_keyname(value_t *valuep, value_t args);
static int function_keycode(value_t *valuep, value_t args);
static int function_screen(value_t *valuep, value_t args);
static int function_scrstat(value_t *valuep, value_t args);
static int function_flush(value_t *valuep, value_t args);
static int function_getcolor(value_t *valuep, value_t args);
static int function_setcolor(value_t *valuep, value_t args);
static int function_getattr(value_t *valuep, value_t args);
static int function_setattr(value_t *valuep, value_t args);
static int function_getpos(value_t *valuep, value_t args);
static int function_move(value_t *valuep, value_t args);
static int function_getch(value_t *valuep, value_t args);
static int function_clear(value_t *valuep, value_t args);
static int function_scroll(value_t *valuep, value_t args);
static int function_line(value_t *valuep, value_t args);
static int function_box(value_t *valuep, value_t args);
static int function_inkey(value_t *valuep, value_t args);
static int function_gscreen(value_t *valuep, value_t args);
static int function_gscrstat(value_t *valuep, value_t args);
static int function_gcolor(value_t *valuep, value_t args);
static int function_grgb(value_t *valuep, value_t args);
static int function_gsync(value_t *valuep, value_t args);
static int function_gflush(value_t *valuep, value_t args);
static int function_ggetsize(value_t *valuep, value_t args);
static int function_ggetoffset(value_t *valuep, value_t args);
static int function_gsetoffset(value_t *valuep, value_t args);
static int function_ggetscale(value_t *valuep, value_t args);
static int function_gsetscale(value_t *valuep, value_t args);
static int function_ggetcolor(value_t *valuep, value_t args);
static int function_gsetcolor(value_t *valuep, value_t args);
static int function_ggetpixel(value_t *valuep, value_t args);
static int function_gclear(value_t *valuep, value_t args);
static int function_gscroll(value_t *valuep, value_t args);
static int function_gdot(value_t *valuep, value_t args);
static int function_gline(value_t *valuep, value_t args);
static int function_gbox(value_t *valuep, value_t args);
static int function_gcircle(value_t *valuep, value_t args);
static int function_gpaint(value_t *valuep, value_t args);
static int function_gcopyimage(value_t *valuep, value_t args);
static int function_ginkey(value_t *valuep, value_t args);
static int function_gkeystat(value_t *valuep, value_t args);
static int function_gmousepos(value_t *valuep, value_t args);
static int function_ginmouse(value_t *valuep, value_t args);
static int function_gmousestat(value_t *valuep, value_t args);
static int function_gmainimage(value_t *valuep, value_t args);
static int function_gallocimage(value_t *valuep, value_t args);
static int function_gfreeimage(value_t *valuep, value_t args);
static int function_gloadimage(value_t *valuep, value_t args);
static int function_gsaveimage(value_t *valuep, value_t args);
static int function_astop(value_t *valuep, value_t args);
static int function_aclear(value_t *valuep, value_t args);
static int function_asound(value_t *valuep, value_t args);
static int function_asndstat(value_t *valuep, value_t args);
static int function_aplaynum(value_t *valuep, value_t args);
static int function_aunitset(value_t *valuep, value_t args);
static int function_aunit(value_t *valuep, value_t args);
static int function_aosc(value_t *valuep, value_t args);
static int function_aaddout(value_t *valuep, value_t args);
static int function_aenvpoint(value_t *valuep, value_t args);
static int function_aenvcopy(value_t *valuep, value_t args);
static int function_afilpoint(value_t *valuep, value_t args);
static int function_afilcopy(value_t *valuep, value_t args);
static int function_aplay(value_t *valuep, value_t args);

static struct function functions_orig[] = {
  { "FARGS"      , "VALUE"  , FUNCTION_TYPE_FARGS      , function_fargs       },
  { "COPY"       , NULL     , FUNCTION_TYPE_COPY       , function_copy        },
  { "TRUE"       , NULL     , FUNCTION_TYPE_TRUE       , function_true        },
  { "FALSE"      , NULL     , FUNCTION_TYPE_FALSE      , function_false       },
  { "COND"       , NULL     , FUNCTION_TYPE_COND       , function_cond        },
  { "AT"         , NULL     , FUNCTION_TYPE_AT         , function_at          },
  { "FPUSH"      , "LOCAL"  , FUNCTION_TYPE_FPUSH      , function_fpush       },
  { "FPOP"       , "RESTORE", FUNCTION_TYPE_FPOP       , function_fpop        },
  { "STACKNUM"   , NULL     , FUNCTION_TYPE_STACKNUM   , function_stacknum    },
  { "FPRINT"     , NULL     , FUNCTION_TYPE_FPRINT     , function_fprint      },
  { "SPRINTF"    , NULL     , FUNCTION_TYPE_SPRINTF    , function_sprintf     },
  { "FPRINTF"    , NULL     , FUNCTION_TYPE_FPRINTF    , function_fprintf     },
  { "FPRINTM"    , NULL     , FUNCTION_TYPE_FPRINTM    , function_fprintm     },
  { "INPUT"      , NULL     , FUNCTION_TYPE_INPUT      , function_input       },
  { "FWAIT"      , NULL     , FUNCTION_TYPE_FWAIT      , function_fwait       },
  { "NOSYSTEM"   , NULL     , FUNCTION_TYPE_NOSYSTEM   , function_nosystem    },
  { "NOSYSCALL"  , NULL     , FUNCTION_TYPE_NOSYSCALL  , function_nosyscall   },
#ifdef NLL_STDIO_FUNCTION
  { "FSTDIN"     , NULL     , FUNCTION_TYPE_FSTDIN     , function_fstdin      },
  { "FSTDOUT"    , NULL     , FUNCTION_TYPE_FSTDOUT    , function_fstdout     },
#endif
#ifdef NLL_SYSTEM_FUNCTION
  { "SYSTEM"     , NULL     , FUNCTION_TYPE_SYSTEM     , function_system      },
#endif
#ifdef NLL_SYSCALL_FUNCTION
  { "SYSCALL"    , NULL     , FUNCTION_TYPE_SYSCALL    , function_syscall     },
#endif
  { "FEVAL"      , NULL     , FUNCTION_TYPE_FEVAL      , function_feval       },
  { "ITOC"       , "CHAR"   , FUNCTION_TYPE_ITOC       , function_itoc        },
  { "CTOI"       , "ASCII"  , FUNCTION_TYPE_CTOI       , function_ctoi        },
  { "ITOA"       , "STR"    , FUNCTION_TYPE_ITOA       , function_itoa        },
  { "ATOI"       , "VAL"    , FUNCTION_TYPE_ATOI       , function_atoi        },
#ifdef NLL_FLOATING_POINT
  { "ITOF"       , NULL     , FUNCTION_TYPE_ITOF       , function_itof        },
  { "FTOI"       , NULL     , FUNCTION_TYPE_FTOI       , function_ftoi        },
#endif
  { "LENGTH"     , "LEN"    , FUNCTION_TYPE_LENGTH     , function_length      },
  { "LEFT"       , NULL     , FUNCTION_TYPE_LEFT       , function_left        },
  { "RIGHT"      , NULL     , FUNCTION_TYPE_RIGHT      , function_right       },
  { "MID"        , NULL     , FUNCTION_TYPE_MID        , function_mid         },
  { "INSTR"      , NULL     , FUNCTION_TYPE_INSTR      , function_instr       },
  { "STRCPY"     , NULL     , FUNCTION_TYPE_STRCPY     , function_strcpy      },
  { "SRAND"      , NULL     , FUNCTION_TYPE_SRAND      , function_srand       },
  { "RAND"       , "RND"    , FUNCTION_TYPE_RAND       , function_rand        },
  { "INT"        , NULL     , FUNCTION_TYPE_INT        , function_int         },
  { "ABS"        , NULL     , FUNCTION_TYPE_ABS        , function_abs         },
  { "SIGN"       , "SGN"    , FUNCTION_TYPE_SIGN       , function_sign        },
  { "TIME"       , NULL     , FUNCTION_TYPE_TIME       , function_time        },
  { "DATE"       , NULL     , FUNCTION_TYPE_DATE       , function_date        },
  { "MEMORY"     , NULL     , FUNCTION_TYPE_MEMORY     , function_memory      },
  { "PEEK"       , NULL     , FUNCTION_TYPE_PEEK       , function_peek        },
  { "POKE"       , NULL     , FUNCTION_TYPE_POKE       , function_poke        },
  { "READ"       , NULL     , FUNCTION_TYPE_READ       , function_read        },
  { "WRITE"      , NULL     , FUNCTION_TYPE_WRITE      , function_write       },
  { "MEMCPY"     , NULL     , FUNCTION_TYPE_MEMCPY     , function_memcpy      },
#ifdef USE_MATHLIB
  { "PI"         , NULL     , FUNCTION_TYPE_PI         , function_pi          },
  { "RADIAN"     , NULL     , FUNCTION_TYPE_RADIAN     , function_radian      },
  { "DEGREE"     , NULL     , FUNCTION_TYPE_DEGREE     , function_degree      },
  { "SIN"        , NULL     , FUNCTION_TYPE_SIN        , function_sin         },
  { "COS"        , NULL     , FUNCTION_TYPE_COS        , function_cos         },
  { "TAN"        , NULL     , FUNCTION_TYPE_TAN        , function_tan         },
  { "ASIN"       , NULL     , FUNCTION_TYPE_ASIN       , function_asin        },
  { "ACOS"       , NULL     , FUNCTION_TYPE_ACOS       , function_acos        },
  { "ATAN"       , NULL     , FUNCTION_TYPE_ATAN       , function_atan        },
  { "SINH"       , NULL     , FUNCTION_TYPE_SINH       , function_sinh        },
  { "COSH"       , NULL     , FUNCTION_TYPE_COSH       , function_cosh        },
  { "TANH"       , NULL     , FUNCTION_TYPE_TANH       , function_tanh        },
  { "EXP"        , NULL     , FUNCTION_TYPE_EXP        , function_exp         },
  { "LOG"        , NULL     , FUNCTION_TYPE_LOG        , function_log         },
  { "LOG10"      , NULL     , FUNCTION_TYPE_LOG10      , function_log10       },
  { "SQRT"       , "SQR"    , FUNCTION_TYPE_SQRT       , function_sqrt        },
  { "ATAN2"      , NULL     , FUNCTION_TYPE_ATAN2      , function_atan2       },
  { "POW"        , NULL     , FUNCTION_TYPE_POW        , function_pow         },
  { "FMOD"       , NULL     , FUNCTION_TYPE_FMOD       , function_fmod        },
  { "INFINITY"   , "INF"    , FUNCTION_TYPE_INFINITY   , function_infinity    },
  { "NAN"        , NULL     , FUNCTION_TYPE_NAN        , function_nan         },
  { "ISFINITE"   , NULL     , FUNCTION_TYPE_ISFINITE   , function_isfinite    },
  { "ISINF"      , NULL     , FUNCTION_TYPE_ISINF      , function_isinf       },
  { "ISNAN"      , NULL     , FUNCTION_TYPE_ISNAN      , function_isnan       },
  { "ISNORMAL"   , NULL     , FUNCTION_TYPE_ISNORMAL   , function_isnormal    },
#endif
  { "FOPEN"      , NULL     , FUNCTION_TYPE_FOPEN      , function_fopen       },
  { "FCLOSE"     , NULL     , FUNCTION_TYPE_FCLOSE     , function_fclose      },
  { "FGETS"      , NULL     , FUNCTION_TYPE_FGETS      , function_fgets       },
  { "FPUTS"      , NULL     , FUNCTION_TYPE_FPUTS      , function_fputs       },
  { "FREAD"      , NULL     , FUNCTION_TYPE_FREAD      , function_fread       },
  { "FWRITE"     , NULL     , FUNCTION_TYPE_FWRITE     , function_fwrite      },
  { "KEYNAME"    , NULL     , FUNCTION_TYPE_KEYNAME    , function_keyname     },
  { "KEYCODE"    , NULL     , FUNCTION_TYPE_KEYCODE    , function_keycode     },
  { "SCREEN"     , NULL     , FUNCTION_TYPE_SCREEN     , function_screen      },
  { "SCRSTAT"    , NULL     , FUNCTION_TYPE_SCRSTAT    , function_scrstat     },
  { "FLUSH"      , NULL     , FUNCTION_TYPE_FLUSH      , function_flush       },
  { "GETCOLOR"   , NULL     , FUNCTION_TYPE_GETCOLOR   , function_getcolor    },
  { "SETCOLOR"   , NULL     , FUNCTION_TYPE_SETCOLOR   , function_setcolor    },
  { "GETATTR"    , NULL     , FUNCTION_TYPE_GETATTR    , function_getattr     },
  { "SETATTR"    , NULL     , FUNCTION_TYPE_SETATTR    , function_setattr     },
  { "GETPOS"     , NULL     , FUNCTION_TYPE_GETPOS     , function_getpos      },
  { "MOVE"       , NULL     , FUNCTION_TYPE_MOVE       , function_move        },
  { "GETCH"      , NULL     , FUNCTION_TYPE_GETCH      , function_getch       },
  { "CLEAR"      , NULL     , FUNCTION_TYPE_CLEAR      , function_clear       },
  { "SCROLL"     , NULL     , FUNCTION_TYPE_SCROLL     , function_scroll      },
  { "LINE"       , NULL     , FUNCTION_TYPE_LINE       , function_line        },
  { "BOX"        , NULL     , FUNCTION_TYPE_BOX        , function_box         },
  { "INKEY"      , "KEY"    , FUNCTION_TYPE_INKEY      , function_inkey       },
  { "GSCREEN"    , NULL     , FUNCTION_TYPE_GSCREEN    , function_gscreen     },
  { "GSCRSTAT"   , NULL     , FUNCTION_TYPE_GSCRSTAT   , function_gscrstat    },
  { "GCOLOR"     , NULL     , FUNCTION_TYPE_GCOLOR     , function_gcolor      },
  { "GRGB"       , NULL     , FUNCTION_TYPE_GRGB       , function_grgb        },
  { "GSYNC"      , NULL     , FUNCTION_TYPE_GSYNC      , function_gsync       },
  { "GFLUSH"     , NULL     , FUNCTION_TYPE_GFLUSH     , function_gflush      },
  { "GGETSIZE"   , NULL     , FUNCTION_TYPE_GGETSIZE   , function_ggetsize    },
  { "GGETOFFSET" , NULL     , FUNCTION_TYPE_GGETOFFSET , function_ggetoffset  },
  { "GSETOFFSET" , NULL     , FUNCTION_TYPE_GSETOFFSET , function_gsetoffset  },
  { "GGETSCALE"  , NULL     , FUNCTION_TYPE_GGETSCALE  , function_ggetscale   },
  { "GSETSCALE"  , NULL     , FUNCTION_TYPE_GSETSCALE  , function_gsetscale   },
  { "GGETCOLOR"  , NULL     , FUNCTION_TYPE_GGETCOLOR  , function_ggetcolor   },
  { "GSETCOLOR"  , NULL     , FUNCTION_TYPE_GSETCOLOR  , function_gsetcolor   },
  { "GGETPIXEL"  , NULL     , FUNCTION_TYPE_GGETPIXEL  , function_ggetpixel   },
  { "GCLEAR"     , NULL     , FUNCTION_TYPE_GCLEAR     , function_gclear      },
  { "GSCROLL"    , NULL     , FUNCTION_TYPE_GSCROLL    , function_gscroll     },
  { "GDOT"       , "GPOINT" , FUNCTION_TYPE_GDOT       , function_gdot        },
  { "GLINE"      , NULL     , FUNCTION_TYPE_GLINE      , function_gline       },
  { "GBOX"       , NULL     , FUNCTION_TYPE_GBOX       , function_gbox        },
  { "GCIRCLE"    , NULL     , FUNCTION_TYPE_GCIRCLE    , function_gcircle     },
  { "GPAINT"     , NULL     , FUNCTION_TYPE_GPAINT     , function_gpaint      },
  { "GCOPYIMAGE" , NULL     , FUNCTION_TYPE_GCOPYIMAGE , function_gcopyimage  },
  { "GINKEY"     , "GKEY"   , FUNCTION_TYPE_GINKEY     , function_ginkey      },
  { "GKEYSTAT"   , NULL     , FUNCTION_TYPE_GKEYSTAT   , function_gkeystat    },
  { "GMOUSEPOS"  , NULL     , FUNCTION_TYPE_GMOUSEPOS  , function_gmousepos   },
  { "GINMOUSE"   , NULL     , FUNCTION_TYPE_GINMOUSE   , function_ginmouse    },
  { "GMOUSESTAT" , NULL     , FUNCTION_TYPE_GMOUSESTAT , function_gmousestat  },
  { "GMAINIMAGE" , NULL     , FUNCTION_TYPE_GMAINIMAGE , function_gmainimage  },
  { "GALLOCIMAGE", NULL     , FUNCTION_TYPE_GALLOCIMAGE, function_gallocimage },
  { "GFREEIMAGE" , NULL     , FUNCTION_TYPE_GFREEIMAGE , function_gfreeimage  },
  { "GLOADIMAGE" , NULL     , FUNCTION_TYPE_GLOADIMAGE , function_gloadimage  },
  { "GSAVEIMAGE" , NULL     , FUNCTION_TYPE_GSAVEIMAGE , function_gsaveimage  },
  { "ASTOP"      , NULL     , FUNCTION_TYPE_ASTOP      , function_astop       },
  { "ACLEAR"     , NULL     , FUNCTION_TYPE_ACLEAR     , function_aclear      },
  { "ASOUND"     , NULL     , FUNCTION_TYPE_ASOUND     , function_asound      },
  { "ASNDSTAT"   , NULL     , FUNCTION_TYPE_ASNDSTAT   , function_asndstat    },
  { "APLAYNUM"   , NULL     , FUNCTION_TYPE_APLAYNUM   , function_aplaynum    },
  { "AUNITSET"   , NULL     , FUNCTION_TYPE_AUNITSET   , function_aunitset    },
  { "AUNIT"      , NULL     , FUNCTION_TYPE_AUNIT      , function_aunit       },
  { "AOSC"       , NULL     , FUNCTION_TYPE_AOSC       , function_aosc        },
  { "AADDOUT"    , NULL     , FUNCTION_TYPE_AADDOUT    , function_aaddout     },
  { "AENVPOINT"  , NULL     , FUNCTION_TYPE_AENVPOINT  , function_aenvpoint   },
  { "AENVCOPY"   , NULL     , FUNCTION_TYPE_AENVCOPY   , function_aenvcopy    },
  { "AFILPOINT"  , NULL     , FUNCTION_TYPE_AFILPOINT  , function_afilpoint   },
  { "AFILCOPY"   , NULL     , FUNCTION_TYPE_AFILCOPY   , function_afilcopy    },
  { "APLAY"      , NULL     , FUNCTION_TYPE_APLAY      , function_aplay       },
  { NULL         , NULL     , FUNCTION_TYPE_NONE       , NULL                 }
};

const char *function_name(function_t function)
{
  return function->name;
}

/* #define NLL_FUNCTION_SHORTNAME */

static int _search(struct function *functions, const char *name, function_t *functionp)
{
  function_t function;
#ifdef NLL_FUNCTION_SHORTNAME
  int len;
#endif

  for (function = functions; function->name; function++) {
    if (!strcmp(function->name, name)) {
      *functionp = function;
      return 0;
    }
  }

  for (function = functions; function->name; function++) {
    if (function->other_name && !strcmp(function->other_name, name)) {
      *functionp = function;
      return 0;
    }
  }

#ifdef NLL_FUNCTION_SHORTNAME
  for (len = 0; name[len]; len++) {
    if ((name[len] == '.') && !name[len + 1]) {
      for (function = functions; function->name; function++) {
	if (!strncmp(function->name, name, len)) {
	  *functionp = function;
	  return 0;
	}
      }
      break;
    }
  }
#endif

  return NLL_ERRCODE_FORMULA_UNKNOWN_FUNCTION;
}

int function_search(const char *name, function_t *functionp)
{
  int r;
  r = _search(functions_orig, name, functionp);
#ifdef NLL_EXTEND_FUNCTION
  if (r < 0)
    r = _search(functions_ext, name, functionp);
#endif
  return r;
}

static int _list(struct function *functions, FILE *fp)
{
  function_t function;

  for (function = functions; function->name; function++) {
    fprintf(fp, "%s", function->name);
    if (function->other_name)
      fprintf(fp, "\t(%s)", function->other_name);
    fprintf(fp, "\n");
  }

  return 0;
}

int function_list(FILE *fp)
{
  _list(functions_orig, fp);
#ifdef NLL_EXTEND_FUNCTION
  _list(functions_ext, fp);
#endif
  return 0;
}

int function_proc(function_t function, value_t *valuep, value_t args)
{
  int r;

  if (value_get_type(args) == VALUE_TYPE_VECTOR) {
    if ((r = value_get_vector(args, &args)) < 0)
      return r;
  }

  return function->function(valuep, args);
}

static int function_fargs(value_t *valuep, value_t args)
{
  int r;

  for (; args; args = args->next) {
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_value(*valuep, args)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_copy(value_t *valuep, value_t args)
{
  int r;

  for (; args; args = args->next) {
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_copy_value(*valuep, args)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_true(value_t *valuep, value_t args)
{
  int r;

  if (!args) {
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_integer(*valuep, 1)) < 0)
      return r;
    valuep = &(*valuep)->next;
  } else {
    for (; args; args = args->next) {
      if ((r = value_extend_value(valuep)) < 0)
	return r;
      if ((r = value_set_integer(*valuep, 1)) < 0)
	return r;
      valuep = &(*valuep)->next;
    }
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_false(value_t *valuep, value_t args)
{
  int r;

  if (!args) {
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_integer(*valuep, 0)) < 0)
      return r;
    valuep = &(*valuep)->next;
  } else {
    for (; args; args = args->next) {
      if ((r = value_extend_value(valuep)) < 0)
	return r;
      if ((r = value_set_integer(*valuep, 0)) < 0)
	return r;
      valuep = &(*valuep)->next;
    }
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_cond(value_t *valuep, value_t args)
{
  int r, n;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_is_true(args)) < 0)
    return r;
  n = r ? 0 : 1;
  args = args->next;

  for (; args && (n > 0); args = args->next, n--)
    ;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if (!args) {
    if ((r = value_clear(*valuep)) < 0)
      return r;
  } else {
    if ((r = value_copy_value(*valuep, args)) < 0)
      return r;
  }
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_at(value_t *valuep, value_t args)
{
  int r;
  integer_t integer = 0;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  args = args->next;

  for (; args && (integer > 0); args = args->next, integer--)
    ;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if (!args) {
    if ((r = value_clear(*valuep)) < 0)
      return r;
  } else {
    if ((r = value_copy_value(*valuep, args)) < 0)
      return r;
  }
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_fpush(value_t *valuep, value_t args)
{
  int r;

  for (; args; args = args->next) {
    if ((r = stack_push(args)) < 0)
      return r;
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_copy_value(*valuep, args)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_fpop(value_t *valuep, value_t args)
{
  int r;

  if (!args) {
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = stack_pop(*valuep)) < 0)
      return r;
    valuep = &(*valuep)->next;
  } else {
    for (; args; args = args->next) {
      if ((r = stack_pop(args)) < 0)
	return r;
      if ((r = value_extend_value(valuep)) < 0)
	return r;
      if ((r = value_copy_value(*valuep, args)) < 0)
	return r;
      valuep = &(*valuep)->next;
    }
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_stacknum(value_t *valuep, value_t args)
{
  int r, n;

  n = stack_getnum();

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, n)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_fprint(value_t *valuep, value_t args)
{
  int r, n = 0;

  value_output_values(args, NULL);

  for (; args; args = args->next, n++)
    ;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, n)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_sprintf(value_t *valuep, value_t args)
{
  int r, n;
  char *format;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &format, NULL)) < 0)
    return r;
  args = args->next;

  n = nll_sprintf(buffer, sizeof(buffer), format, args);
  if (n < 0)
    strcpy(buffer, "");

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, buffer, -1)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, n)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_fprintf(value_t *valuep, value_t args)
{
  int r, n;
  char *format;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &format, NULL)) < 0)
    return r;
  args = args->next;

  n = nll_sprintf(buffer, sizeof(buffer), format, args);
  if (n > 0)
    framebuf_puts(buffer);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, n)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_fprintm(value_t *valuep, value_t args)
{
  int r, result;
  integer_t x = -1, y = -1, fgcolor = -1, bgcolor = -1;
  value_t arg;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  arg = args;
  args = args->next;

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &fgcolor)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &bgcolor)) < 0)
      ;
  }

  result = framebuf_move_cursor(x, y);
  if (fgcolor >= 0) framebuf_setfgcolor(fgcolor);
  if (bgcolor >= 0) framebuf_setbgcolor(bgcolor);
  value_output_value(arg);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_input(value_t *valuep, value_t args)
{
  int r, len;
  char *string;
  char buffer[STRING_MAXLEN + 1];

  framebuf_enable(0);

  for (; args; args = args->next) {
    if (value_get_string(args, &string, NULL) < 0)
      continue;
    fprintf(nll_stdout, "%s", string);
  }
  fflush(nll_stdout);

  if (!fgets(buffer, sizeof(buffer), nll_stdin))
    buffer[0] = '\0';

  len = strlen(buffer);
  while ((len > 0) && (buffer[len - 1] == '\n'))
    buffer[--len] = '\0';

  framebuf_enable(1);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, buffer, -1)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_fwait(value_t *valuep, value_t args)
{
  int r, usec = 1, sec = 0;
  integer_t integer;

  for (; args; args = args->next) {
    if (value_get_integer(args, &integer) < 0)
      ;
    else {
      if (usec)
	nll_usleep(integer * usec);
      if (sec)
	nll_sleep(integer * sec);
    }
    usec *= 100;
    sec  *= 100;
    if (usec == 1000000) {
      usec = 0;
      sec  = 1;
    }
  }

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, 0)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_nosystem(value_t *valuep, value_t args)
{
  int r, result;

  result = nll_nosystem_set();

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_nosyscall(value_t *valuep, value_t args)
{
  int r, result;

  result = nll_nosyscall_set();

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

#ifdef NLL_STDIO_FUNCTION
static int function_fstdin(value_t *valuep, value_t args)
{
  int r, result;
  char *filename;
  FILE *fp;

  if (nll_is_nosystem())
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &filename, NULL)) < 0)
    return r;

  if (!filename) {
    result = (nll_stdin_restore() < 0) ? -1 : 0;
  } else {
    fp = fopen(filename, "r");
    if (fp)
      result = (nll_stdin_save(fp) < 0) ? -1 : 0;
    else
      result = -1;
  }

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_fstdout(value_t *valuep, value_t args)
{
  int r, result;
  char *filename;
  FILE *fp;

  if (nll_is_nosystem())
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &filename, NULL)) < 0)
    return r;

  if (!filename) {
    result = (nll_stdout_restore() < 0) ? -1 : 0;
  } else {
    fp = fopen(filename, "w");
    if (fp)
      result = (nll_stdout_save(fp) < 0) ? -1 : 0;
    else
      result = -1;
  }

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}
#endif

#ifdef NLL_SYSTEM_FUNCTION
static int function_system(value_t *valuep, value_t args)
{
  int r, result;
  char *command;

  if (nll_is_nosystem())
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &command, NULL)) < 0)
    return r;

#ifndef USE_NLLIBC
  result = system(command);
#else
  result = -1;
#endif

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}
#endif

#ifdef NLL_SYSCALL_FUNCTION
#define SYSCALL_ARG_MAXNUM 8

static int syscall_call(int syscall_no, int argc, long *args)
{
  int r = -1;

  switch (argc) {
#ifndef WIN32
#ifndef USE_NLLIBC
  case 0: r = syscall(syscall_no                                    ); break;
  case 1: r = syscall(syscall_no, args[0]                           ); break;
  case 2: r = syscall(syscall_no, args[0], args[1]                  ); break;
  case 3: r = syscall(syscall_no, args[0], args[1], args[2]         ); break;
  case 4: r = syscall(syscall_no, args[0], args[1], args[2], args[3]); break;
  case 5: r = syscall(syscall_no, args[0], args[1], args[2], args[3],
		      args[4]                                       ); break;
  case 6: r = syscall(syscall_no, args[0], args[1], args[2], args[3],
		      args[4], args[5]                              ); break;
  case 7: r = syscall(syscall_no, args[0], args[1], args[2], args[3],
		      args[4], args[5], args[6]                     ); break;
  case 8: r = syscall(syscall_no, args[0], args[1], args[2], args[3],
		      args[4], args[5], args[6], args[7]            ); break;
#endif
#endif
  default:
    break;
  }

  return r;
}

static int function_syscall(value_t *valuep, value_t args)
{
  int r, result, argc = 0;
  integer_t syscall_no, integer;
  char *string;
  area_t area;
  long syscall_args[SYSCALL_ARG_MAXNUM];

  if (nll_is_nosystem())
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
  if (nll_is_nosyscall())
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &syscall_no)) < 0)
    return r;
  args = args->next;

  for (; args; args = args->next) {
    if ((r = value_get_integer(args, &integer)) < 0) {
      if ((r = value_get_string(args, &string, NULL)) < 0) {
	if ((r = value_get_area(args, &area)) < 0) {
	  return r;
	} else {
	  syscall_args[argc] = (long)area_get_buffer(area);
	}
      } else {
	syscall_args[argc] = (long)string;
      }
    } else {
      syscall_args[argc] = (long)integer;
    }
    argc++;
    if (argc == SYSCALL_ARG_MAXNUM)
      break;
  }

  result = syscall_call(syscall_no, argc, syscall_args);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}
#endif

static int function_feval(value_t *valuep, value_t args)
{
  int r;
  char *string;
  value_t retval;
  line_t line;
  struct spot s;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &string, NULL)) < 0)
    return r;

  if ((r = line_set(-1, -1, string, &line)) < 0)
    return r;
  if (!line)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  if ((r = nll_parse(line)) < 0)
    goto err;

  s.line = line;
  s.command = line->commands;
  if ((r = nll_exec(&s, NULL, POSITION_TYPE_EVAL, &retval, NULL)) < 0)
    goto err;

  for (; retval; retval = retval->next) {
    if ((r = value_extend_value(valuep)) < 0)
      goto err;
    if ((r = value_copy_value(*valuep, retval)) < 0)
      goto err;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    goto err;

  if ((r = line_free(line)) < 0)
    return r;

  return 0;

err:
  if (line)
    line_free(line);
  return r;
}

static int function_itoc(value_t *valuep, value_t args)
{
  int r;
  integer_t integer;
  char str[2];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_integer(args, &integer)) < 0)
      return r;
    str[0] = integer;
    str[1] = '\0';
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_string(*valuep, str, -1)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_ctoi(value_t *valuep, value_t args)
{
  int r;
  integer_t integer;
  char *string;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_string(args, &string, NULL)) < 0)
      return r;
    integer = *(unsigned char *)string;
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_integer(*valuep, integer)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_itoa(value_t *valuep, value_t args)
{
  int r;
  integer_t integer;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_integer(args, &integer)) < 0)
      return r;
    sprintf(buffer, "%ld", integer);
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_string(*valuep, buffer, -1)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_atoi(value_t *valuep, value_t args)
{
  int r;
  integer_t integer;
  char *string;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_string(args, &string, NULL)) < 0)
      return r;
    integer = (int)strtol(string, NULL, 0);
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_integer(*valuep, integer)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

#ifdef NLL_FLOATING_POINT
static int function_itof(value_t *valuep, value_t args)
{
  int r;
  integer_t integer;
  double f;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_integer(args, &integer)) < 0)
      return r;
    f = (double)integer;
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_float(*valuep, f)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_ftoi(value_t *valuep, value_t args)
{
  int r;
  integer_t integer;
  double f;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_float(args, &f)) < 0)
      return r;
    integer = (int)f;
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_integer(*valuep, integer)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}
#endif

static int function_length(value_t *valuep, value_t args)
{
  int r, length;
  char *string;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_string(args, &string, &length)) < 0)
      return r;
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_integer(*valuep, length)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_left(value_t *valuep, value_t args)
{
  int r, i, length;
  integer_t integer = 1;
  char *s, *p;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s, &length)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &integer)) < 0)
      ;
  }

  p = buffer;
  for (i = 0; (i < length) && (i < integer); i++) {
    *(p++) = *(s++);
  }
  *p = '\0';

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, buffer, i)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_right(value_t *valuep, value_t args)
{
  int r, i, length;
  integer_t integer = 1;
  char *s, *p;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s, &length)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &integer)) < 0)
      ;
  }

  s += length - integer;
  p = buffer;
  for (i = 0; i < integer; i++) {
    *(p++) = *(s++);
  }
  *p = '\0';

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, buffer, i)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_mid(value_t *valuep, value_t args)
{
  int r, i, length;
  integer_t integer0 = 1, integer1 = -1;
  char *s, *p;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s, &length)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &integer0)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &integer1)) < 0)
      ;
  }

  s += integer0;
  p = buffer;
  for (i = 0; (integer0 + i < length) && ((integer1 < 0) || (i < integer1)); i++) {
    *(p++) = *(s++);
  }
  *p = '\0';

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, buffer, i)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_instr(value_t *valuep, value_t args)
{
  int r, i, n, length0, length1;
  char *s0, *s1;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s0, &length0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s1, &length1)) < 0)
    return r;

  n = -1;
  for (i = 0; i <= length0 - length1; i++) {
    if (!memcmp(s0 + i, s1, length1)) {
      n = i;
      break;
    }
  }

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, n)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_strcpy(value_t *valuep, value_t args)
{
  int r, length0, length1;
  integer_t offset = -1;
  char *s0, *s1;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s0, &length0)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s1, &length1)) < 0)
    return r;

  if (offset < 0)
    offset = length0;

  if (offset + length1 > STRING_MAXLEN)
    return NLL_ERRCODE_STRING_TOO_LONG;

  memcpy(buffer, s0, length0);
  memcpy(buffer + offset, s1, length1);

  if (offset + length1 > length0)
    length0 = offset + length1;

  buffer[length0] = '\0';

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, buffer, length0)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static unsigned int seed = 1;

static unsigned int _rand(void)
{
  seed ^= seed << 13;
  seed ^= seed >> 17;
  seed ^= seed <<  5;
  return seed;
}

static int function_srand(value_t *valuep, value_t args)
{
  int r;
  integer_t s;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &s)) < 0)
    return r;

  seed = !s ? time(NULL) : s;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, seed)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_rand(value_t *valuep, value_t args)
{
  int r, n;
  integer_t range = 0;

  if (!args) {
    _rand();
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_integer(*valuep, seed)) < 0)
      return r;
    valuep = &(*valuep)->next;
  } else {
    for (; args; args = args->next) {
      if ((r = value_get_integer(args, &range)) < 0)
	return r;
      n = range;
      _rand();
      if ((r = value_extend_value(valuep)) < 0)
	return r;
      if ((r = value_set_integer(*valuep, (n > 0) ? (seed % n) : seed)) < 0)
	return r;
      valuep = &(*valuep)->next;
    }
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_int(value_t *valuep, value_t args)
{
  int r;
  integer_t integer;
#ifdef NLL_FLOATING_POINT
  double f;
#endif

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_integer(args, &integer)) < 0) {
#ifdef NLL_FLOATING_POINT
      if ((r = value_get_float(args, &f)) < 0)
	return r;
      integer = f;
      while (integer > f)
	integer--;
#else
      return r;
#endif
    }
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_integer(*valuep, integer)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_abs(value_t *valuep, value_t args)
{
  int r, is_float;
  integer_t integer;
#ifdef NLL_FLOATING_POINT
  double f;
#endif

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    is_float = 0;
    if ((r = value_get_integer(args, &integer)) < 0) {
#ifdef NLL_FLOATING_POINT
      if ((r = value_get_float(args, &f)) < 0)
	return r;
      f = (f < 0.0) ? -f : f;
      is_float = 1;
#else
      return r;
#endif
    } else {
      integer = (integer < 0) ? -integer : integer;
    }
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if (is_float) {
#ifdef NLL_FLOATING_POINT
      if ((r = value_set_float(*valuep, f)) < 0)
	return r;
#else
      return NLL_ERRCODE_VALUE_INVALID_TYPE;
#endif
    } else {
      if ((r = value_set_integer(*valuep, integer)) < 0)
	return r;
    }
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_sign(value_t *valuep, value_t args)
{
  int r, is_float;
  integer_t integer;
#ifdef NLL_FLOATING_POINT
  double f;
#endif

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    is_float = 0;
    if ((r = value_get_integer(args, &integer)) < 0) {
#ifdef NLL_FLOATING_POINT
      if ((r = value_get_float(args, &f)) < 0)
	return r;
      if      (f < 0.0) f = -1.0;
      else if (f > 0.0) f =  1.0;
      else              f =  0.0;
      is_float = 1;
#else
      return r;
#endif
    } else {
      if      (integer < 0) integer = -1;
      else if (integer > 0) integer =  1;
      else                  integer =  0;
    }
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if (is_float) {
#ifdef NLL_FLOATING_POINT
      if ((r = value_set_float(*valuep, f)) < 0)
	return r;
#else
      return NLL_ERRCODE_VALUE_INVALID_TYPE;
#endif
    } else {
      if ((r = value_set_integer(*valuep, integer)) < 0)
	return r;
    }
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_time(value_t *valuep, value_t args)
{
  int r, n;
  time_t t;

  t = time(NULL);
  n = t;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, n)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_date(value_t *valuep, value_t args)
{
  int r;
#ifndef USE_NLLIBC
  time_t t;
#endif
  char *s;

#ifndef USE_NLLIBC
  t = time(NULL);
  s = ctime(&t);
#else
  s = "Cannot get date.";
#endif

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, s, -1)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_memory(value_t *valuep, value_t args)
{
  int r;
  integer_t integer;
  area_t area;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  if ((r = area_alloc(&area, integer)) < 0)
    return r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_area(*valuep, area)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_peek(value_t *valuep, value_t args)
{
  int r;
  integer_t offset = 0, size = -1, integer;
  area_t area;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
  }

  if ((r = area_read_integer(area, offset, &integer, size)) < 0)
    return r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, integer)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_poke(value_t *valuep, value_t args)
{
  int r;
  integer_t offset = 0, size = -1, integer;
  area_t area;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
    args = args->next;
  }

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  if ((r = area_write_integer(area, offset, integer, size)) < 0)
    return r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, integer)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_read(value_t *valuep, value_t args)
{
  int r;
  integer_t offset = 0, size = -1;
  area_t area;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
  }

  if (size < 0)
    size = area_get_size(area) - offset;

  if (size < 0)
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  if (size > STRING_MAXLEN)
    return NLL_ERRCODE_STRING_TOO_LONG;

  if ((r = area_read(area, offset, buffer, size)) < 0)
    return r;
  buffer[size] = '\0';

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, buffer, size)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_write(value_t *valuep, value_t args)
{
  int r, length;
  integer_t offset = 0, size = -1;
  area_t area;
  char *string;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
    args = args->next;
  }

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &string, &length)) < 0)
    return r;

  if (size < 0)
    size = length;

  if (size > length)
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  if ((r = area_write(area, offset, string, size)) < 0)
    return r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, size)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_memcpy(value_t *valuep, value_t args)
{
  int r;
  integer_t doffset = 0, soffset = 0, size = -1;
  area_t darea, sarea;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &darea)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &doffset)) < 0)
      ;
    args = args->next;
  }

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &sarea)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &soffset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
  }

  if (size < 0)
    size = area_get_size(sarea) - soffset;

  if (size < 0)
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  if ((r = area_copy_area(darea, doffset, sarea, soffset, size)) < 0)
    return r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, size)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

#ifdef USE_MATHLIB
static int function_mathfunc(function_type_t type, value_t *valuep, value_t args)
{
#ifndef NLL_FLOATING_POINT
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  double f;

  if (!args) {
    switch (type) {
    case FUNCTION_TYPE_PI: f = M_PI; break;
    default:
      return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
    }
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_float(*valuep, f)) < 0)
      return r;
    valuep = &(*valuep)->next;
  } else {
    for (; args; args = args->next) {
      if ((r = value_get_float_integer(args, &f)) < 0)
	return r;
      switch (type) {
      case FUNCTION_TYPE_PI:     f = M_PI * f;                 break;
      case FUNCTION_TYPE_RADIAN: f = (2.0 * M_PI) / 360.0 * f; break;
      case FUNCTION_TYPE_DEGREE: f = 360.0 / (2.0 * M_PI) * f; break;
      case FUNCTION_TYPE_SIN:    f = sin(f);                   break;
      case FUNCTION_TYPE_COS:    f = cos(f);                   break;
      case FUNCTION_TYPE_TAN:    f = tan(f);                   break;
      case FUNCTION_TYPE_ASIN:   f = asin(f);                  break;
      case FUNCTION_TYPE_ACOS:   f = acos(f);                  break;
      case FUNCTION_TYPE_ATAN:   f = atan(f);                  break;
      case FUNCTION_TYPE_SINH:   f = sinh(f);                  break;
      case FUNCTION_TYPE_COSH:   f = cosh(f);                  break;
      case FUNCTION_TYPE_TANH:   f = tanh(f);                  break;
      case FUNCTION_TYPE_EXP:    f = exp(f);                   break;
      case FUNCTION_TYPE_LOG:    f = log(f);                   break;
      case FUNCTION_TYPE_LOG10:  f = log10(f);                 break;
      case FUNCTION_TYPE_SQRT:   f = sqrt(f);                  break;
      default:
	return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
      }
      if ((r = value_extend_value(valuep)) < 0)
	return r;
      if ((r = value_set_float(*valuep, f)) < 0)
	return r;
      valuep = &(*valuep)->next;
    }
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_pi(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_PI, valuep, args);
}

static int function_radian(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_RADIAN, valuep, args);
}

static int function_degree(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_DEGREE, valuep, args);
}

static int function_sin(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_SIN, valuep, args);
}

static int function_cos(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_COS, valuep, args);
}

static int function_tan(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_TAN, valuep, args);
}

static int function_asin(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_ASIN, valuep, args);
}

static int function_acos(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_ACOS, valuep, args);
}

static int function_atan(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_ATAN, valuep, args);
}

static int function_sinh(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_SINH, valuep, args);
}

static int function_cosh(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_COSH, valuep, args);
}

static int function_tanh(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_TANH, valuep, args);
}

static int function_exp(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_EXP, valuep, args);
}

static int function_log(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_LOG, valuep, args);
}

static int function_log10(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_LOG10, valuep, args);
}

static int function_sqrt(value_t *valuep, value_t args)
{
  return function_mathfunc(FUNCTION_TYPE_SQRT, valuep, args);
}

static int function_mathfunc2(function_type_t type, value_t *valuep, value_t args)
{
#ifndef NLL_FLOATING_POINT
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  double f0, f1;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_float_integer(args, &f0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_float_integer(args, &f1)) < 0)
    return r;

  switch (type) {
  case FUNCTION_TYPE_ATAN2: f0 = atan2(f0, f1); break;
  case FUNCTION_TYPE_POW:   f0 = pow(f0, f1);   break;
  case FUNCTION_TYPE_FMOD:  f0 = fmod(f0, f1);  break;
  default:
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
  }

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_float(*valuep, f0)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_atan2(value_t *valuep, value_t args)
{
  return function_mathfunc2(FUNCTION_TYPE_ATAN2, valuep, args);
}

static int function_pow(value_t *valuep, value_t args)
{
  return function_mathfunc2(FUNCTION_TYPE_POW, valuep, args);
}

static int function_fmod(value_t *valuep, value_t args)
{
  return function_mathfunc2(FUNCTION_TYPE_FMOD, valuep, args);
}

static int function_mathvalue(function_type_t type, value_t *valuep, value_t args)
{
#ifndef NLL_FLOATING_POINT
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  double f;

  if (!args) {
    switch (type) {
#ifdef USE_MATHBUILTIN
    case FUNCTION_TYPE_INFINITY: f = INFINITY; break;
    case FUNCTION_TYPE_NAN:      f = NAN;      break;
#endif
    default:
      return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
    }
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_float(*valuep, f)) < 0)
      return r;
    valuep = &(*valuep)->next;
  } else {
    for (; args; args = args->next) {
      switch (type) {
#ifdef USE_MATHBUILTIN
      case FUNCTION_TYPE_INFINITY: f = INFINITY; break;
      case FUNCTION_TYPE_NAN:      f = NAN;      break;
#endif
      default:
	return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
      }
      if ((r = value_extend_value(valuep)) < 0)
	return r;
      if ((r = value_set_float(*valuep, f)) < 0)
	return r;
      valuep = &(*valuep)->next;
    }
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_infinity(value_t *valuep, value_t args)
{
  return function_mathvalue(FUNCTION_TYPE_INFINITY, valuep, args);
}

static int function_nan(value_t *valuep, value_t args)
{
  return function_mathvalue(FUNCTION_TYPE_NAN, valuep, args);
}

static int function_mathcheck(function_type_t type, value_t *valuep, value_t args)
{
#ifndef NLL_FLOATING_POINT
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, is = 0;
  double f;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  for (; args; args = args->next) {
    if ((r = value_get_float_integer(args, &f)) < 0)
      return r;
    switch (type) {
#ifdef USE_MATHBUILTIN
    case FUNCTION_TYPE_ISFINITE: is = isfinite(f) ? 1 : 0; break;
    case FUNCTION_TYPE_ISINF   : is = isinf(f)    ? 1 : 0; break;
    case FUNCTION_TYPE_ISNAN   : is = isnan(f)    ? 1 : 0; break;
    case FUNCTION_TYPE_ISNORMAL: is = isnormal(f) ? 1 : 0; break;
#endif
    default:
      return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
    }
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_integer(*valuep, is)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_isfinite(value_t *valuep, value_t args)
{
  return function_mathcheck(FUNCTION_TYPE_ISFINITE, valuep, args);
}

static int function_isinf(value_t *valuep, value_t args)
{
  return function_mathcheck(FUNCTION_TYPE_ISINF, valuep, args);
}

static int function_isnan(value_t *valuep, value_t args)
{
  return function_mathcheck(FUNCTION_TYPE_ISNAN, valuep, args);
}

static int function_isnormal(value_t *valuep, value_t args)
{
  return function_mathcheck(FUNCTION_TYPE_ISNORMAL, valuep, args);
}
#endif

static int function_fopen(value_t *valuep, value_t args)
{
  int r;
  integer_t integer;
  char *filename, *mode = "r";
  FILE *fp;

  if (nll_is_nosystem())
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &filename, NULL)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_string(args, &mode, NULL)) < 0)
      ;
  }

  fp = fopen(filename, mode);
  integer = fp ? (integer_t)fp : 0;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, integer)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_fclose(value_t *valuep, value_t args)
{
  int r, result;
  integer_t integer;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  fp = (FILE *)integer;
  result = fclose(fp);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_fgets(value_t *valuep, value_t args)
{
  int r;
  integer_t integer;
  char buffer[STRING_MAXLEN + 1];
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  fp = integer ? (FILE *)integer : nll_stdin;
  if (!fgets(buffer, sizeof(buffer), fp))
    buffer[0] = '\0';

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, buffer, -1)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_fputs(value_t *valuep, value_t args)
{
  int r, length;
  integer_t integer;
  char *string;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &string, &length)) < 0)
    return r;

  fp = integer ? (FILE *)integer : nll_stdout;
  if (!fputs(string, fp))
    length = -1;
  fflush(fp);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, length)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_fread(value_t *valuep, value_t args)
{
  int r;
  area_t area;
  integer_t integer, offset = 0, size = -1;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
  }

  if (size < 0)
    size = area_get_size(area) - offset;

  if (size < 0)
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  if ((offset < 0) || (offset + size > area_get_size(area)))
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  fp = integer ? (FILE *)integer : nll_stdin;
  size = fread(area_get_buffer(area) + offset, 1, size, fp);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, size)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_fwrite(value_t *valuep, value_t args)
{
  int r;
  area_t area;
  integer_t integer, offset = 0, size = -1;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &offset)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &size)) < 0)
      ;
  }

  if (size < 0)
    size = area_get_size(area) - offset;

  if (size < 0)
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  if ((offset < 0) || (offset + size > area_get_size(area)))
    return NLL_ERRCODE_AREA_OUT_OF_RANGE;

  fp = integer ? (FILE *)integer : nll_stdout;
  size = fwrite(area_get_buffer(area) + offset, 1, size, fp);
  fflush(fp);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, size)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_keyname(value_t *valuep, value_t args)
{
  int r;
  integer_t keycode;
  char *keyname;

  for (; args; args = args->next) {
    if ((r = value_get_integer(args, &keycode)) < 0)
      continue;
    keyname = keyname_get(keycode);
    if (!keyname) keyname = "";
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_string(*valuep, keyname, -1)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_keycode(value_t *valuep, value_t args)
{
  int r;
  char *keyname;
  key_code_t keycode;

  for (; args; args = args->next) {
    if ((r = value_get_string(args, &keyname, NULL)) < 0)
      continue;
    keycode = keycode_get(keyname);
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_integer(*valuep, keycode)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_screen(value_t *valuep, value_t args)
{
  int r, result;
  integer_t flags = -1, width = -1, height = -1;

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &width)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &height)) < 0)
      ;
  }

  result = framebuf_setscreen(flags, width, height);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_scrstat(value_t *valuep, value_t args)
{
  int r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, framebuf_get_screen_flags())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, framebuf_get_screen_width())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, framebuf_get_screen_height())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_flush(value_t *valuep, value_t args)
{
  int r, result;
  integer_t flags = 0;

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
  }

  result = framebuf_flush(flags);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_getcolor(value_t *valuep, value_t args)
{
  int r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, framebuf_get_cursor_fgcolor())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, framebuf_get_cursor_bgcolor())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_setcolor(value_t *valuep, value_t args)
{
  int r, result;
  integer_t fgcolor = -1, bgcolor = -1;

  if (args) {
    if ((r = value_get_integer(args, &fgcolor)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &bgcolor)) < 0)
      ;
  }

  result = framebuf_setfgcolor(fgcolor);
  framebuf_setbgcolor(bgcolor);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_getattr(value_t *valuep, value_t args)
{
  int r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, framebuf_get_cursor_effect())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_setattr(value_t *valuep, value_t args)
{
  int r, result;
  integer_t effect = -1;

  if (args) {
    if ((r = value_get_integer(args, &effect)) < 0)
      ;
  }

  result = framebuf_seteffect(effect);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_getpos(value_t *valuep, value_t args)
{
  int r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, framebuf_get_cursor_x())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, framebuf_get_cursor_y())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_move(value_t *valuep, value_t args)
{
  int r, result;
  integer_t x = -1, y = -1, dx = 0, dy = 0;

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &dx)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &dy)) < 0)
      ;
  }

  result = framebuf_move_cursor(x, y);

  if      (dx < 0) framebuf_move_left(-dx);
  else if (dx > 0) framebuf_move_right(dx);

  if      (dy < 0) framebuf_move_up(-dy);
  else if (dy > 0) framebuf_move_down(dy);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_getch(value_t *valuep, value_t args)
{
  int r;
  integer_t x = -1, y = -1;
  int ch;
  char s[2];

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
  }

  ch = framebuf_getch(x, y);

  s[0] = ch;
  s[1] = '\0';

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, s, -1)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_clear(value_t *valuep, value_t args)
{
  int r, result;
  integer_t ch = ' ', flags = 0;

  if (args) {
    if ((r = value_get_integer(args, &ch)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
  }

  result = framebuf_clear(ch, flags);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_scroll(value_t *valuep, value_t args)
{
  int r, result;
  integer_t x = 0, y = 0, ch = -1;

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &ch)) < 0)
      ;
  }

  result = framebuf_scroll(x, y, ch);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_line(value_t *valuep, value_t args)
{
  int r, result;
  integer_t x0, y0, x1, y1;
  char *string = "";

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x1)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y1)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_string(args, &string, NULL)) < 0)
      ;
  }

  result = framebuf_line(x0, y0, x1, y1, string);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_box(value_t *valuep, value_t args)
{
  int r, result;
  integer_t x0, y0, x1, y1, flags = 0;
  char *string = "";

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x1)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y1)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_string(args, &string, NULL)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
  }

  result = framebuf_box(x0, y0, x1, y1, string, flags);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_inkey(value_t *valuep, value_t args)
{
  int r;
  integer_t integer;
  char key[2], *p;

  integer = framebuf_getkey();

  switch (integer) {
  case KEY_CODE_END:
  case KEY_CODE_UP:
  case KEY_CODE_DOWN:
  case KEY_CODE_RIGHT:
  case KEY_CODE_LEFT:
  case KEY_CODE_PAGEUP:
  case KEY_CODE_PAGEDOWN:
    p = keyname_get(integer);
    break;
  default:
    key[0] = integer;
    key[1] = '\0';
    p = key;
    break;
  }

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, p, -1)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
}

static int function_gscreen(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t flags = -1, width = -1, height = -1, index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &width)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &height)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_setscreen(flags, width, height, index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gscrstat(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, graphic_get_screen_flags())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, graphic_get_screen_width())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, graphic_get_screen_height())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, graphic_get_screen_image())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gcolor(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t red = 0, green = 0, blue = 0;
  int color;

  if (args) {
    if ((r = value_get_integer(args, &red)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &green)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &blue)) < 0)
      ;
  }

  color = graphic_color(red, green, blue);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, color)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_grgb(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t color = 0;
  int red, green, blue;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
  }

  graphic_rgb(color, &red, &green, &blue);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, red)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, green)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, blue)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gsync(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;

  result = graphic_sync();

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gflush(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t flags = 0;

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
  }

  result = graphic_flush(flags);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_ggetsize(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t index = NLL_G_MAINIMG;
  int width, height;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  width  = graphic_getwidth(index);
  height = graphic_getheight(index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, width)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, height)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_ggetoffset(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t index = NLL_G_MAINIMG;
  int x = 0, y = 0;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  graphic_getoffset(index, &x, &y);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, x)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, y)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gsetoffset(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t x = 0, y = 0, index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_setoffset(index, x, y);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_ggetscale(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t index = NLL_G_MAINIMG;
  int mulx = 0, muly = 0, divx = 0, divy = 0;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  graphic_getscale(index, &mulx, &muly, &divx, &divy);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, mulx)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, muly)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, divx)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, divy)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gsetscale(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t mulx = 0, muly = 0, divx = 0, divy = 0, index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &mulx)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &muly)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &divx)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &divy)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_setscale(index, mulx, muly, divx, divy);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_ggetcolor(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_getcolor(index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gsetcolor(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t color = 0, index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_setcolor(index, color);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_ggetpixel(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  integer_t x = -1, y = -1, index = NLL_G_MAINIMG;
  int color;

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  color = graphic_getpixel(index, x, y);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, color)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gclear(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t color = 0, flags = 0, index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_clear(index, color, flags);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gscroll(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t x = 0, y = 0, color = 0, flags = 0, index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_scroll(index, x, y, color, flags);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gdot(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t x, y, color = -1, flags = 0, index = NLL_G_MAINIMG;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_dot(index, x, y, color, flags);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gline(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t x0, y0, x1, y1, color = -1, flags = 0, index = NLL_G_MAINIMG;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x1)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y1)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_line(index, x0, y0, x1, y1, color, flags);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gbox(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t x0, y0, x1, y1, color = -1, flags = 0, index = NLL_G_MAINIMG;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x1)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y1)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_box(index, x0, y0, x1, y1, color, flags);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gcircle(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t x, y, radius, color = -1, flags = 0, index = NLL_G_MAINIMG;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &radius)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_circle(index, x, y, radius, color, flags);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gpaint(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t x, y, border = -1, color = -1, flags = 0, index = NLL_G_MAINIMG;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &x)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer_float(args, &y)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &border)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &color)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_paint(index, x, y, border, color, flags);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gcopyimage(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t x0 = 0, y0 = 0, source = NLL_G_MAINIMG, x = 0, y = 0, width = 0, height = 0, transparent = -1, flags = 0, index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer_float(args, &x0)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y0)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &source)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &width)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &height)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &transparent)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_image(index, x0, y0, source, x, y, width, height, transparent, flags);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_ginkey(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, keycode;
  char *keyname;

  keycode = graphic_getkeybuf();
  keyname = keyname_get(keycode);
  if (!keyname) keyname = "";

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, keyname, -1)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gkeystat(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  char *keyname;
  int keycode, s;

  for (; args; args = args->next) {
    if ((r = value_get_string(args, &keyname, NULL)) < 0)
      continue;
    keycode = keycode_get(keyname);
    s = graphic_getkeystatus(keycode);
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_integer(*valuep, s)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gmousepos(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, graphic_getmousex())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, graphic_getmousey())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_ginmouse(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, mousecode, x, y;
  char *mousename;

  mousecode = graphic_getmousebuf(&x, &y);
  mousename = mousename_get(mousecode);
  if (!mousename) {
    mousename = "";
    x = -1;
    y = -1;
  }

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_string(*valuep, mousename, -1)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, x)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, y)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gmousestat(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;
  char *mousename;
  int mousecode, s;

  for (; args; args = args->next) {
    if ((r = value_get_string(args, &mousename, NULL)) < 0)
      continue;
    mousecode = mousecode_get(mousename);
    s = graphic_getmousestatus(mousecode);
    if ((r = value_extend_value(valuep)) < 0)
      return r;
    if ((r = value_set_integer(*valuep, s)) < 0)
      return r;
    valuep = &(*valuep)->next;
  }

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gmainimage(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_mainimage(index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gallocimage(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t width = 0, height = 0;

  if (args) {
    if ((r = value_get_integer_float(args, &width)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &height)) < 0)
      ;
  }

  result = graphic_allocimage(width, height);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gfreeimage(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t index = NLL_G_MAINIMG;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
  }

  result = graphic_freeimage(index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gloadimage(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result, width = 0, height = 0;
  char *filename;

  if (nll_is_nosystem())
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &filename, NULL)) < 0)
    return r;

  result = graphic_loadimage(filename, &width, &height);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, width)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, height)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_gsaveimage(value_t *valuep, value_t args)
{
#ifndef USE_GRAPHIC
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t source = NLL_G_MAINIMG, x = 0, y = 0, width = 0, height = 0;
  char *filename;

  if (nll_is_nosystem())
    return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &filename, NULL)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &source)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &x)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &y)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &width)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer_float(args, &height)) < 0)
      ;
  }

  result = graphic_saveimage(filename, source, x, y, width, height);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_astop(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;

  result = audio_stop();

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_aclear(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t aunitset_index = -1;

  if (args) {
    if ((r = value_get_integer(args, &aunitset_index)) < 0)
      ;
  }

  result = audio_clear(aunitset_index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_asound(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t samplefreq = 0;

  if (args) {
    if ((r = value_get_integer(args, &samplefreq)) < 0)
      ;
  }

  result = audio_setsamplefreq(samplefreq);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_asndstat(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, audio_getsamplefreq())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_aplaynum(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r;

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, audio_playnum())) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_aunitset(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t aunitset_index = NLL_A_NULLUNITSET;

  if (args) {
    if ((r = value_get_integer(args, &aunitset_index)) < 0)
      ;
  }

  result = audio_unitset(aunitset_index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_aunit(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t output = NLL_A_OUTPUT, amp = AUDIO_PER_NUM, flags = 0, aunitset_index = NLL_A_NULLUNITSET;

  if (args) {
    if ((r = value_get_integer(args, &output)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &amp)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &aunitset_index)) < 0)
      ;
  }

  result = audio_unit(output, amp, flags, aunitset_index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_aosc(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t index = 1, amp = AUDIO_PER_NUM, freq = AUDIO_PER_NUM, phase = 0, type = NLL_A_SIN, mod = NLL_A_PHASE, flags = 0, aunitset_index = NLL_A_NULLUNITSET;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &amp)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &freq)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &phase)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &type)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &mod)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &aunitset_index)) < 0)
      ;
  }

  result = audio_osc(index, amp, freq, phase, type, mod, flags, aunitset_index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_aaddout(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t index = 1, output = NLL_A_OUTPUT, amp = AUDIO_PER_NUM, flags = 0, aunitset_index = NLL_A_NULLUNITSET;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &output)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &amp)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &aunitset_index)) < 0)
      ;
  }

  result = audio_addout(index, output, amp, flags, aunitset_index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_aenvpoint(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t index = 1, n = -1, step = 0, amp = AUDIO_PER_NUM, flags = 0, aunitset_index = NLL_A_NULLUNITSET;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &n)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &step)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &amp)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &aunitset_index)) < 0)
      ;
  }

  result = audio_envpoint(index, n, step, amp, flags, aunitset_index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_aenvcopy(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t index = 1, source = 1, flags = 0, aunitset_index = NLL_A_NULLUNITSET;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &source)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &aunitset_index)) < 0)
      ;
  }

  result = audio_envcopy(index, source, flags, aunitset_index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_afilpoint(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t index = 1, type = NLL_A_FILFW, n = -1, amp = AUDIO_PER_NUM, flags = 0, aunitset_index = NLL_A_NULLUNITSET;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &type)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &n)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &amp)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &aunitset_index)) < 0)
      ;
  }

  result = audio_filpoint(index, type, n, amp, flags, aunitset_index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_afilcopy(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  integer_t index = 1, source = 1, flags = 0, aunitset_index = NLL_A_NULLUNITSET;

  if (args) {
    if ((r = value_get_integer(args, &index)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &source)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &aunitset_index)) < 0)
      ;
  }

  result = audio_filcopy(index, source, flags, aunitset_index);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}

static int function_aplay(value_t *valuep, value_t args)
{
#ifndef USE_AUDIO
  return NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION;
#else
  int r, result;
  char *string, *filename = NULL;
  integer_t flags = NLL_A_FLUSH, aunitset_index = NLL_A_NULLUNITSET;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &string, NULL)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &flags)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &aunitset_index)) < 0)
      ;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_string(args, &filename, NULL)) < 0)
      ;
  }

  result = audio_play(string, flags, aunitset_index, filename);

  if ((r = value_extend_value(valuep)) < 0)
    return r;
  if ((r = value_set_integer(*valuep, result)) < 0)
    return r;
  valuep = &(*valuep)->next;

  if ((r = value_shrink_value(valuep)) < 0)
    return r;

  return 0;
#endif
}
