#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "nlltypes.h"
#include "nlllib.h"
#include "value.h"
#include "stack.h"

static struct stack_frame stack[STACK_MAXNUM];
static int stack_num = 0;

int stack_check(void)
{
  stack_frame_t stack_frame;
  int i;

  if (stack_num)
    return NLL_ERRCODE_STACK_NOT_EMPTY;

  for (i = 0; i < STACK_MAXNUM; i++) {
    stack_frame = &stack[i];
    if (stack_frame->value)
      return NLL_ERRCODE_STACK_NOT_EMPTY;
  }

  return 0;
}

int stack_init(void)
{
  stack_frame_t stack_frame;
  int r, i;

  memset(stack, 0, sizeof(stack));

  for (i = 0; i < STACK_MAXNUM; i++) {
    stack_frame = &stack[i];
    if ((r = value_alloc(&stack_frame->value)) < 0)
      return r;
  }

  stack_num = 0;

  return 0;
}

int stack_getnum(void)
{
  return stack_num;
}

static int _push(value_t value)
{
  stack_frame_t stack_frame;
  int r;

  if (stack_num >= STACK_MAXNUM)
    return NLL_ERRCODE_STACK_STACK_OVER;

  stack_frame = &stack[stack_num];

  if (value) {
    if ((r = value_copy_value(stack_frame->value, value)) < 0)
      return r;
  }

  stack_num++;

  return 0;
}

static int _pop(value_t value)
{
  stack_frame_t stack_frame;
  int r;

  if (stack_num == 0)
    return NLL_ERRCODE_STACK_NOT_FOUND;

  stack_frame = &stack[stack_num - 1];

  stack_num--;

  if (value) {
    if ((r = value_copy_value(value, stack_frame->value)) < 0)
      return r;
  }

  if ((r = value_clear(stack_frame->value)) < 0)
    return r;

  return 0;
}

int stack_push(value_t value)
{
  return _push(value);
}

int stack_pop(value_t value)
{
  return _pop(value);
}

int stack_clear(void)
{
  stack_frame_t stack_frame;
  int r, i;

  for (i = 0; i < STACK_MAXNUM; i++) {
    stack_frame = &stack[i];
    if ((r = value_free(stack_frame->value)) < 0)
      return r;
    stack_frame->value = NULL;
  }

  stack_num = 0;

  return 0;
}

int stack_clean(void)
{
  int r;

  while (stack_num > 0) {
    if ((r = stack_pop(NULL)) < 0)
      return r;
  }

  return 0;
}

int stack_dump(FILE *fp)
{
  stack_frame_t stack_frame;
  int i;

  nll_wait_output(fp);
  fprintf(fp, "\n-- Stack dump --\n");

  for (i = 0; i < stack_num; i++) {
    nll_wait_output(fp);
    stack_frame = &stack[i];
    fprintf(fp, "[%d]\tValue:", i);
    value_dump(fp, stack_frame->value);
    fprintf(fp, "\n");
  }

  fflush(fp);

  return 0;
}
